import os
from typing import Any, Dict, List, Optional, Union

from dotenv import load_dotenv

from ai_atlas_nexus.blocks.inference.base import InferenceEngine
from ai_atlas_nexus.blocks.inference.params import (
    InferenceEngineCredentials,
    OpenAIChatCompletionMessageParam,
    TextGenerationInferenceOutput,
    WMLInferenceEngineParams,
)
from ai_atlas_nexus.blocks.inference.postprocessing import postprocess
from ai_atlas_nexus.metadata_base import InferenceEngineType
from ai_atlas_nexus.toolkit.job_utils import run_parallel
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)

# load .env file to environment
load_dotenv()


class WMLInferenceEngine(InferenceEngine):

    _inference_engine_type = InferenceEngineType.WML
    _inference_engine_parameter_class = WMLInferenceEngineParams

    def prepare_credentials(
        self, credentials: Union[Dict, InferenceEngineCredentials]
    ) -> InferenceEngineCredentials:
        api_key = credentials.get(
            "api_key", os.environ.get(f"{self._inference_engine_type}_API_KEY", None)
        )
        assert api_key, (
            f"Error while trying to create {self._inference_engine_type}. "
            f"Please set the env variable: '{self._inference_engine_type}_API_KEY' or pass api_key to credentials."
        )

        api_url = credentials.get(
            "api_url", os.environ.get(f"{self._inference_engine_type}_API_URL", None)
        )
        assert api_url, (
            f"Error while trying to create {self._inference_engine_type}. "
            f"Please set the env variable: '{self._inference_engine_type}_API_URL' or pass api_url to credentials."
        )

        space_id = credentials.get(
            "space_id",
            os.environ.get(f"{self._inference_engine_type}_SPACE_ID", None),
        )

        project_id = credentials.get(
            "project_id",
            os.environ.get(f"{self._inference_engine_type}_PROJECT_ID", None),
        )

        if not (space_id or project_id):
            raise ValueError(
                "Error while trying to create {self._inference_engine_type} inference engine. "
                "Either 'space_id' or 'project_id' need to be specified when using WMLInferenceEngine. "
                f"Please set the env variable: '{self._inference_engine_type}_SPACE_ID'/'{self._inference_engine_type}_PROJECT_ID' "
                "or pass space_id/project_id to credentials.",
            )
        elif space_id and project_id:
            logger.warning(
                "Either 'space_id' or 'project_id' need to be "
                "specified, however, both were found. 'WMLInferenceEngine' "
                "will use space_id by default."
            )

        return InferenceEngineCredentials(
            api_key=api_key, api_url=api_url, space_id=space_id, project_id=project_id
        )

    def create_client(self, credentials: InferenceEngineCredentials):
        from ibm_watsonx_ai import APIClient
        from ibm_watsonx_ai.foundation_models import ModelInference

        client = APIClient(
            {
                "url": credentials["api_url"],
                "apikey": credentials["api_key"],
            }
        )
        if credentials["space_id"]:
            client.set.default_space(credentials["space_id"])
        else:
            client.set.default_project(credentials["project_id"])

        # self.parameters.update(
        #     {"response_format": self._create_schema_format(response_format)}
        # )
        return ModelInference(
            model_id=self.model_name_or_path, api_client=client, params=self.parameters
        )

    def ping(self):
        # ping is handled by the APIClient class internally.
        pass

    @postprocess
    def generate(
        self,
        prompts: List[str],
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> List[TextGenerationInferenceOutput]:
        responses = []
        for response in self.client.generate(
            prompt=prompts,
            params=self.parameters,
            concurrency_limit=self.concurrency_limit,
        ):
            responses.append(self._prepare_generation_output(response))

        return responses

    def _prepare_generation_output(
        self, response
    ) -> List[TextGenerationInferenceOutput]:
        return TextGenerationInferenceOutput(
            prediction=response["results"][0]["generated_text"],
            input_tokens=response["results"][0]["input_token_count"],
            output_tokens=response["results"][0]["generated_token_count"],
            stop_reason=response["results"][0]["stop_reason"],
            model_name_or_path=self.model_name_or_path,
            inference_engine=str(self._inference_engine_type),
        )

    @postprocess
    def chat(
        self,
        messages: Union[
            List[OpenAIChatCompletionMessageParam],
            List[str],
        ],
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> TextGenerationInferenceOutput:

        def chat_response(messages):
            response = self.client.chat(
                messages=self._to_openai_format(messages),
                params=self.parameters,
            )
            return self._prepare_chat_output(response)

        return run_parallel(
            chat_response,
            messages,
            f"Inferring with {self._inference_engine_type}",
            self.concurrency_limit,
            verbose=verbose,
        )

    def _prepare_chat_output(self, response) -> List[TextGenerationInferenceOutput]:
        print(response["choices"][0]["message"]["content"])
        return TextGenerationInferenceOutput(
            prediction=response["choices"][0]["message"]["content"],
            input_tokens=response["usage"]["prompt_tokens"],
            output_tokens=response["usage"]["completion_tokens"],
            stop_reason=response["choices"][0]["finish_reason"],
            model_name_or_path=self.model_name_or_path,
            inference_engine=str(self._inference_engine_type),
        )

    def _create_schema_format(self, response_format):
        if response_format:
            return {
                "type": "json_schema",
                "json_schema": {
                    "name": "RITS_schema",
                    "schema": response_format,
                },
            }
        else:
            return None



import os
from typing import Dict, List, Optional, Union

from dotenv import load_dotenv

from ai_atlas_nexus.blocks.inference.base import InferenceEngine
from ai_atlas_nexus.blocks.inference.params import (
    InferenceEngineCredentials,
    OllamaInferenceEngineParams,
    OpenAIChatCompletionMessageParam,
    TextGenerationInferenceOutput,
)
from ai_atlas_nexus.blocks.inference.postprocessing import postprocess
from ai_atlas_nexus.metadata_base import InferenceEngineType
from ai_atlas_nexus.toolkit.job_utils import run_parallel
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)

# load .env file to environment
load_dotenv()


class OllamaInferenceEngine(InferenceEngine):

    _inference_engine_type = InferenceEngineType.OLLAMA
    _inference_engine_parameter_class = OllamaInferenceEngineParams

    def prepare_credentials(
        self, credentials: Union[Dict, InferenceEngineCredentials]
    ) -> InferenceEngineCredentials:
        api_url = credentials.get(
            "api_url",
            os.environ.get(f"{self._inference_engine_type}_API_URL", None),
        )
        assert api_url, (
            f"Error while trying to run {self._inference_engine_type}. "
            f"Please pass api_url to credentials or set the env variable: '{self._inference_engine_type}_API_URL'"
        )

        if api_url:
            logger.info(
                f"{self._inference_engine_type} inference engine will execute requests on the server at {api_url}."
            )

        return InferenceEngineCredentials(api_url=api_url)

    def create_client(self, credentials):
        from ollama import Client

        return Client(host=credentials["api_url"])

    def ping(self):
        if self.model_name_or_path not in [
            model.model for model in self.client.list().models
        ]:
            raise Exception(
                f"Model `{self.model_name_or_path}` not found. Please download it first."
            )

    def is_thinking_supported(self):
        if "thinking" in self.client.show(self.model_name_or_path).capabilities:
            return True
        else:
            raise Exception(
                f"`Model {self.model_name_or_path}` does not support thinking. Please pass `think=False` or use another model."
            )

    @postprocess
    def generate(
        self,
        prompts: List[str],
        response_format=None,
        postprocessors=None,
        verbose=True,
        **kwargs,
    ) -> List[TextGenerationInferenceOutput]:
        def generate_text(prompt: str):
            response = self.client.generate(
                model=self.model_name_or_path,
                prompt=prompt,
                format=response_format,
                options=self.parameters,  # https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values
                think=self.think,
                **kwargs,
            )
            return self._prepare_prediction_output(response)

        return run_parallel(
            generate_text,
            prompts,
            f"Inferring with {self._inference_engine_type}",
            self.concurrency_limit,
            verbose=verbose,
        )

    @postprocess
    def chat(
        self,
        messages: Union[
            List[OpenAIChatCompletionMessageParam],
            List[str],
        ],
        tools=None,
        response_format=None,
        postprocessors=None,
        verbose=True,
        **kwargs,
    ) -> List[TextGenerationInferenceOutput]:

        def chat_response(messages):
            response = self.client.chat(
                model=self.model_name_or_path,
                messages=self._to_openai_format(messages),
                tools=tools,
                format=response_format,
                options=self.parameters,  # https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values
                think=self.think,
                **kwargs,
            )
            return self._prepare_prediction_output(response)

        return run_parallel(
            chat_response,
            messages,
            f"Inferring with {self._inference_engine_type}",
            self.concurrency_limit,
            verbose=verbose,
        )

    def _prepare_prediction_output(self, response):
        return TextGenerationInferenceOutput(
            prediction=(
                response.message.content if hasattr(response, "message") else response.response
            ),
            input_tokens=response.prompt_eval_count,
            output_tokens=response.eval_count,
            stop_reason=response.done_reason,
            thinking=response.thinking if hasattr(response, "thinking") else None,
            model_name_or_path=self.model_name_or_path,
            inference_engine=str(self._inference_engine_type),
        )



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import (
    Container,
    Risk,
    RiskTaxonomy,
)
from ai_atlas_nexus.blocks.inference.base import InferenceEngine
from ai_atlas_nexus.blocks.prompt_builder import (
    FewShotPromptBuilder,
    ZeroShotPromptBuilder,
)
from ai_atlas_nexus.data import load_resource
from ai_atlas_nexus.toolkit.logging import configure_logger
from ai_atlas_nexus.toolkit.validator import validate


LOGGER = configure_logger(__name__)


RISK_IDENTIFICATION_COT = load_resource("risk_generation_cot.json")
RISK_IDENTIFICATION_COT_SCHEMA = load_resource("risk_generation_cot_schema.json")


class RiskDetector(ABC):

    def __init__(
        self,
        risks: List[Risk],
        inference_engine: InferenceEngine,
        cot_examples: Optional[Dict[str, List]] = None,
        max_risk: Optional[int] = None,
    ):
        self.inference_engine = inference_engine
        self._risks = risks
        self._examples = cot_examples

        # Validate format of user-provided `cot_examples` if available
        if cot_examples is not None and (
            errors := validate(cot_examples, RISK_IDENTIFICATION_COT_SCHEMA)
        ):
            raise Exception(
                f"The format of `cot_examples` is incorrect. {errors}. Please refer to the example template provided "
                f"at src/ai_atlas_nexus/data/templates/risk_generation_cot.json"
            )

        # Set prompt builder based on whether the CoT examples are available.
        if self._examples is None:
            self.prompt_builder = ZeroShotPromptBuilder
        else:
            self.prompt_builder = FewShotPromptBuilder

        self.max_risk = max_risk

    def get_risks_by_taxonomy_id(
        self, ontology: Container, taxonomy_id: Optional[str] = None
    ):
        taxonomies = list(
            filter(
                lambda taxonomy: taxonomy.id == taxonomy_id,
                ontology.taxonomies,
            )
        )

        if len(taxonomies) > 0:
            taxonomy: RiskTaxonomy = taxonomies[0]
            LOGGER.info(
                f"Selected taxonomy is {str(taxonomy.name)}. For more info: {taxonomy.url}"
            )

            return list(
                filter(
                    lambda risk: risk.isDefinedByTaxonomy == taxonomy.id,
                    ontology.risks,
                )
            )
        else:
            raise Exception(
                f"Risk Taxonomy: {taxonomy_id} not found. Available taxonomies: "
                f"{[taxonomy.id for taxonomy in ontology.taxonomies]}"
            )

    @abstractmethod
    def detect(self, usecases: List[str]) -> List[List[Risk]]:
        raise NotImplementedError



import json
import os
from importlib.metadata import version
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from jinja2 import Template
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import YAMLDumper
from sssom_schema import Mapping


# workaround for txtai
os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
os.environ["OMP_NUM_THREADS"] = "1"

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import (
    Action,
    Adapter,
    AiEval,
    BenchmarkMetadataCard,
    Dataset,
    Documentation,
    LLMIntrinsic,
    LLMQuestionPolicy,
    Policy,
    Principle,
    Risk,
    RiskControl,
    RiskIncident,
    RiskTaxonomy,
    Rule,
    Stakeholder,
)
from ai_atlas_nexus.blocks.inference import InferenceEngine
from ai_atlas_nexus.blocks.prompt_builder import (
    FewShotPromptBuilder,
    ZeroShotPromptBuilder,
)
from ai_atlas_nexus.blocks.prompt_response_schema import (
    DOMAIN_TYPE_SCHEMA,
    LIST_OF_STR_SCHEMA,
    QUESTIONNAIRE_OUTPUT_SCHEMA,
)
from ai_atlas_nexus.blocks.prompt_templates import (
    AI_TASKS_TEMPLATE,
    QUESTIONNAIRE_COT_TEMPLATE,
)
from ai_atlas_nexus.blocks.risk_categorization.severity import RiskSeverityCategorizer
from ai_atlas_nexus.blocks.risk_detector import GenericRiskDetector
from ai_atlas_nexus.blocks.risk_explorer import RiskExplorer
from ai_atlas_nexus.blocks.risk_mapping import RiskMapper
from ai_atlas_nexus.data import load_resource
from ai_atlas_nexus.extension import Extension
from ai_atlas_nexus.metadata_base import MappingMethod
from ai_atlas_nexus.toolkit.data_utils import load_yamls_to_container
from ai_atlas_nexus.toolkit.error_utils import type_check, value_check
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)
RISK_IDENTIFICATION_COT = load_resource("risk_generation_cot.json")


class AIAtlasNexus:
    """A AIAtlasNexus object"""

    # Load the schema
    directory = os.path.dirname(os.path.abspath(__file__))
    fn = os.path.join(
        directory,
        "ai_risk_ontology/schema/ai-risk-ontology.yaml",
    )
    schema_view = SchemaView(yaml.safe_load(fn))

    def __init__(self, base_dir: str = None):
        """Create a new AIAtlasNexus object

        Args:
            base_dir: str
                (Optional) add an alternative source of date
        """
        if base_dir is not None:
            if type(base_dir) != str:
                raise ValueError(
                    "Base directory must be a string",
                    base_dir,
                )
            if not os.path.isdir(base_dir):
                logger.error(
                    f"Directory %s does not exist.",
                    base_dir,
                )
                raise FileNotFoundError(
                    "Base directory is not found",
                    base_dir,
                )

        ontology = load_yamls_to_container(base_dir)
        self._ontology = ontology
        self._risk_explorer = RiskExplorer(ontology)
        logger.info(
            f"Created AIAtlasNexus instance. Base_dir: %s",
            base_dir,
        )

    def export(cls, export_path):
        """Export AIAtlasNexus configuration to file.

        Args:
            export_path: str
                The path to the directory where the artifact will be exported to.

        """
        if not os.path.isdir(export_path):
            logger.error(
                f"Directory %s does not exist.",
                export_path,
            )
            raise FileNotFoundError(
                "Export directory is not found",
                export_path,
            )

        export_file_path = os.path.join(export_path, "ai-risk-ontology.yaml")

        with open(
            export_file_path,
            "+tw",
            encoding="utf-8",
        ) as output_file:
            print(
                YAMLDumper().dumps(cls._ontology),
                file=output_file,
            )
            output_file.close()

    @classmethod
    def get_schema(cls):
        """Get schema

        Returns:
            schema
        """
        return cls.schema_view

    @classmethod
    def get_version(cls):
        """Get library version

        Returns:
            dict: Version number
        """
        response = {"version": version("ai_atlas_nexus")}
        return response

    def get_all_risks(cls, taxonomy=None):
        """Get all risk definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Risk]
                Result containing a list of AI risks
        """
        type_check(
            "<RANEACF44A7E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        risk_instances = cls._risk_explorer.get_all_risks(taxonomy)
        return risk_instances

    def get_risk(
        cls,
        tag=None,
        id=None,
        name=None,
        taxonomy=None,
    ):
        """Get risk definition from the LinkML, filtered by risk atlas id, tag, name

        Args:
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Risk
                Result containing a list of AI risks
        """
        type_check(
            "<RAND62C1B3AE>",
            str,
            allow_none=True,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN14D4D967E>",
            tag or id or name,
            "Please provide tag, id, or name",
        )

        risk: Risk | None = cls._risk_explorer.get_risk(
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        return risk

    def get_related_risks(
        cls,
        risk=None,
        tag=None,
        id=None,
        name=None,
        taxonomy=None,
    ):
        """Get related risks from the LinkML, filtered by risk id, tag, or name

        Args:
            risk: (Optional) Risk
                The risk
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy
        Returns:
            List[str]
                Result containing a list of AI risk IDs
        """
        type_check(
            "<RAN283B72CAE>",
            Risk,
            allow_none=True,
            risk=risk,
        )
        type_check(
            "<RANC9FDCC45E>",
            str,
            allow_none=True,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN0748ECB7E>",
            risk or tag or id or name,
            "Please provide tag, id, or name",
        )

        related_risk_instances = cls._risk_explorer.get_related_risks(
            risk=risk,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        return related_risk_instances

    def get_related_actions(
        cls,
        risk=None,
        tag=None,
        id=None,
        name=None,
        taxonomy=None,
    ):
        """Get actions for a risk definition from the LinkML.  The risk is identified by risk id, tag, or name

        Args:
            risk: (Optional) Risk
                The risk
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Risk
                Result containing a list of AI actions
        """
        type_check(
            "<RANEDB39EABE>",
            Risk,
            allow_none=True,
            risk=risk,
        )
        type_check(
            "<RANC49E332BE>",
            str,
            allow_none=True,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN7154EE0FE>",
            risk or tag or id or name,
            "Please provide risk, tag, id, or name",
        )

        actions = cls._risk_explorer.get_related_actions(
            risk=risk,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        return actions

    def get_all_actions(cls, taxonomy=None):
        """Get all action definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Action]
                Result containing a list of AI actions
        """
        type_check(
            "<RAN1C9A35ADE>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        action_instances: list[Action] = cls._risk_explorer.get_all_actions(taxonomy)
        return action_instances

    def get_action_by_id(cls, id, taxonomy=None):
        """Get an action definition from the LinkML, filtered by action id

        Args:
            id: str
                The string id identifying the action
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Action
                Result containing an action
        """
        type_check(
            "<RAN66203B1FE>",
            str,
            allow_none=False,
            id=id,
        )
        type_check(
            "<RAN869039B6E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        action: Action | None = cls._risk_explorer.get_action_by_id(id=id)
        return action

    def get_related_risk_controls(
        cls,
        risk=None,
        tag=None,
        id=None,
        name=None,
        taxonomy=None,
    ):
        """Get related risk controls for a risk definition from the LinkML.  The risk is identified by risk id, tag, or name

        Args:
            risk: (Optional) Risk
                The risk
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Risk
                Result containing a list of AI actions
        """
        type_check(
            "<RAN4E03158FE>",
            Risk,
            allow_none=True,
            risk=risk,
        )
        type_check(
            "<RAN55784808E>",
            str,
            allow_none=True,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN5DCADF94E>",
            risk or tag or id or name,
            "Please provide risk, tag, id, or name",
        )

        risk_controls = cls._risk_explorer.get_related_risk_controls(
            risk=risk,
            tag=tag,
            id=id,
            name=name,
            taxonomy=taxonomy,
        )
        return risk_controls

    def get_all_risk_controls(cls, taxonomy=None):
        """Get all risk control definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[RiskControl]
                Result containing a list of RiskControls
        """
        type_check(
            "<RAN129A1692E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        risk_control_instances: list[RiskControl] = (
            cls._risk_explorer.get_all_risk_controls(taxonomy)
        )
        return risk_control_instances

    def get_risk_control(cls, id=None, taxonomy=None):
        """Get an action definition from the LinkML, filtered by risk control id

        Args:
            id: str
                The string id identifying the risk control
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Action
                Result containing a risk control.
        """
        type_check(
            "<RAN9785FFE3E>",
            str,
            allow_none=False,
            id=id,
        )
        type_check(
            "<RAN5A157049E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        risk_control: RiskControl | None = cls._risk_explorer.get_risk_control(id=id)
        return risk_control

    def identify_risks_from_usecases(
        cls,
        usecases: List[str],
        inference_engine: InferenceEngine,
        taxonomy: Optional[str] = None,
        cot_examples: Optional[Dict[str, List]] = None,
        max_risk: Optional[int] = None,
        zero_shot_only: bool = False,
    ) -> List[List[Risk]]:
        """Identify potential risks from a usecase description

        Args:
            usecases (List[str]):
                A List of strings describing AI usecases
            inference_engine (InferenceEngine):
                An LLM inference engine to infer risks from the usecases.
            taxonomy (str, optional):
                The string label for a taxonomy. If not specified, the system will use "ibm-risk-atlas" as the default taxonomy.
            cot_examples (Dict[str, List], optional):
                If the user wants to improve risk identification via a Few-shot approach, `cot_examples` can be
                provided with the desired taxonomy as key. Please follow the example template at src/ai_atlas_nexus/data/templates/risk_generation_cot.json.
                If the `cot_examples` is omitted, the API default to a Zero-Shot approach.
            max_risk (int, optional):
                The maximum number of risks to extract. Pass None to allow the inference engine to determine the number of risks. Defaults to None.
            zero_shot_only (bool): If enabled, this flag allows the system to perform Zero Shot Risk identification, and the field `cot_examples` will be ignored.
        Returns:
            List[List[Risk]]:
                Result containing a list of risks
        """
        type_check(
            "<RANE023314BE>",
            List,
            allow_none=False,
            usecases=usecases,
        )
        type_check(
            "<RANE023314BE>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check(
            "<RANB72CAE6EE>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )
        type_check(
            "<RAN80975498E>",
            int,
            allow_none=True,
            max_risk=max_risk,
        )
        value_check(
            "<RAN4717CF18E>",
            all([isinstance(usecase, str) for usecase in usecases]),
            "Usecases must be a list of string.",
        )

        # if not providing taxonomy, set to IBM AI risk atlas
        if taxonomy is None:
            logger.warning(
                f"<RAN47375G12W>",
                f"Taxonomy was not provided, defaulting to ibm-risk-atlas.",
            )
        set_taxonomy = taxonomy or "ibm-risk-atlas"

        processed_examples = None
        if zero_shot_only:
            logger.info(
                f"The `zero_shot_only` flag is enabled. The system will use the Zero shot method. Any provided `cot_examples` will be disregarded.",
            )
        else:
            # For the given taxonomy type, check if the user has provided 'cot_examples'. If not,
            # retrieve the default cot examples from the master. If no examples exist in the master,
            # set it as None.
            processed_examples = (
                cot_examples and cot_examples.get(set_taxonomy, None)
            ) or RISK_IDENTIFICATION_COT.get(set_taxonomy, None)
            if processed_examples is None:
                logger.warning(
                    f"<RAN47275F12W> Chain of Thought (CoT) examples were not provided, or do not exist in the master for this taxonomy. The API will use the Zero shot method. To improve the accuracy of risk identification, please provide CoT examples in `cot_examples` when calling this API. You may also consider raising an issue to permanently add these examples to the AI Atlas Nexus master."
                )

        if set_taxonomy == "ibm-attack-risk-atlas":
            risks = [
                risk
                for risk in cls._risk_explorer.get_all_risks("ibm-risk-atlas")
                if risk.tag.endswith("-attack")
            ]
        else:
            risks = cls._risk_explorer.get_all_risks(set_taxonomy)

        risk_detector = GenericRiskDetector(
            risks=risks,
            inference_engine=inference_engine,
            cot_examples=processed_examples,
            max_risk=max_risk,
        )

        return risk_detector.detect(usecases)

    def get_all_taxonomies(cls):
        """Get all taxonomy definitions from the LinkML

        Returns:
            List[RiskTaxonomy]
                Result containing a list of AI Risk taxonomies
        """
        taxonomy_instances: list[RiskTaxonomy] = cls._risk_explorer.get_all_taxonomies()
        return taxonomy_instances

    def get_taxonomy_by_id(cls, id):
        """Get taxonomy definitions from the LinkML filtered by taxonomy id

        Args:
            id: str
                The string id for a taxonomy

        Returns:
            RiskTaxonomy
                An AI Risk taxonomy
        """
        type_check(
            "<RANBFB574E3E>",
            str,
            allow_none=False,
            id=id,
        )

        taxonomy: RiskTaxonomy | None = cls._risk_explorer.get_taxonomy_by_id(id)
        return taxonomy

    def generate_zero_shot_risk_questionnaire_output(
        cls,
        usecase: str,
        risk_questionnaire: List[Dict[str, str]],
        inference_engine: InferenceEngine,
        verbose=True,
    ):
        """Get prediction using the zero shot approach.

        Args:
            usecase (str): A string describing an AI usecase
            risk_questionnaire: List[Dict[str, str]]: A risk questionnaire
                Check example below.
                ```
                [
                    "In which environment is the system used?",
                ]
                ```
            inference_engine (InferenceEngine):
                An LLM inference engine to predict the output based on the given use case.

        Returns:
            List[str]: List of LLM predictions.
        """
        type_check(
            "<RANF7EFFADAE>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check(
            "<RANB9FDEA04E>",
            str,
            allow_none=False,
            usecase=usecase,
        )
        type_check(
            "<RANF7256EC3E>",
            List,
            allow_none=False,
            questions=risk_questionnaire,
        )
        value_check(
            "<RANC49F00D3E>",
            inference_engine and risk_questionnaire,
            "Please provide questions and inference_engine",
        )

        # Extract only questions
        risk_questionnaire = [
            question_data["question"] for question_data in risk_questionnaire
        ]

        # Prepare zero shots inference prompts
        prompts = [
            ZeroShotPromptBuilder(
                QUESTIONNAIRE_COT_TEMPLATE,
            ).build(usecase=usecase, question=question)
            for question in risk_questionnaire
        ]

        # Invoke inference service
        return inference_engine.generate(
            prompts,
            response_format=QUESTIONNAIRE_OUTPUT_SCHEMA,
            postprocessors=["json_object"],
            verbose=verbose,
        )

    def generate_few_shot_risk_questionnaire_output(
        cls,
        usecase: str,
        risk_questionnaire: List[Dict[str, Any]],
        inference_engine: InferenceEngine,
        verbose=True,
    ):
        """Get prediction using the few shot (Chain of Thought) examples.

        Args:
            usecase (str): A string describing an AI usecase
            risk_questionnaire (List[Dict]): Chain of Thought data for risk questionnaire.
                Each question is associated with a list of example intents and
                corresponding answers. Check example JSON below.
                ```
                [
                    {
                        "question": "In which environment is the system used?",
                        "examples": [
                            "intent": "Find patterns in healthcare insurance claims",
                            "answer": "Insurance Claims Processing or Risk Management or Data Analytics",
                            "explanation": "The system might be used by an insurance company's claims processing department to analyze and identify patterns in healthcare insurance claims."
                        ]
                    }
                ]
            inference_engine (InferenceEngine):
                An LLM inference engine to predict the output based on the given use case.
            filter_cot_data_by (Dict[str, str]):
                A dictionary to filter CoT examples with key as CoT field and value as filter string.
                ```

        Returns:
            List[str]: List of LLM predictions.
        """
        type_check(
            "<RAN19989483E>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check(
            "<RAN17812927E>",
            str,
            allow_none=False,
            usecase=usecase,
        )
        type_check(
            "<RAN46376875E>",
            List,
            allow_none=False,
            questions=risk_questionnaire,
        )
        value_check(
            "<RAN59638961E>",
            inference_engine and risk_questionnaire,
            "Please provide risk_questionnaire_cot and inference_engine",
        )

        assert (
            risk_questionnaire and len(risk_questionnaire) > 0
        ), "`Chain of Thought (risk_questionnaire_cot)` data cannot be None or empty."

        # Prepare few shots inference prompts from CoT Data
        prompts = [
            FewShotPromptBuilder(QUESTIONNAIRE_COT_TEMPLATE).build(
                cot_examples=question_data["cot_examples"],
                usecase=usecase,
                question=question_data["question"],
            )
            for question_data in risk_questionnaire
        ]

        # Invoke inference service
        return inference_engine.generate(
            prompts,
            response_format=QUESTIONNAIRE_OUTPUT_SCHEMA,
            postprocessors=["json_object"],
            verbose=verbose,
        )

    def identify_ai_tasks_from_usecases(
        cls, usecases: List[str], inference_engine: InferenceEngine, verbose=True
    ) -> List[List[str]]:
        """Identify potential risks from a usecase description

        Args:
            usecases (List[str]):
                A List of strings describing AI usecases
            inference_engine (InferenceEngine):
                An LLM inference engine to identify AI tasks from usecases.

        Returns:
            List[List[str]]:
                Result containing a list of AI tasks
        """
        type_check(
            "<RAN3B9CD886E>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check("<RAN4CDA6852E>", List, allow_none=False, usecases=usecases)
        value_check(
            "<RAN0E435F50E>",
            inference_engine and usecases,
            "Please provide usecases and inference_engine",
        )

        # Load HF tasks from the template dir
        hf_ai_tasks = load_resource("hf_ai_tasks.json")

        # Populate schema items
        json_schema = dict(LIST_OF_STR_SCHEMA)
        json_schema["items"]["enum"] = [task["task_label"] for task in hf_ai_tasks]

        # Invoke inference service
        return inference_engine.generate(
            prompts=[
                Template(AI_TASKS_TEMPLATE).render(
                    usecase=usecase, hf_ai_tasks=hf_ai_tasks, limit=len(hf_ai_tasks)
                )
                for usecase in usecases
            ],
            response_format=json_schema,
            postprocessors=["list_of_str"],
            verbose=verbose,
        )

    def generate_proposed_mappings(
        cls,
        new_risks: List[Risk],
        existing_risks: List[Risk],
        inference_engine: InferenceEngine,
        new_prefix: str,
        mapping_method: MappingMethod = MappingMethod.SEMANTIC,
    ) -> List[Mapping]:
        """Identify mappings between a new set of risks and risks that exist in the Risk Atlas

        Args:
            new_risks: List[Risk]
                A new set of risks
            existing_risks: List[Risk]
                Secondary list, this should be the list of existing risks in RAN
            inference_engine: (Optional)Union[InferenceEngine | None]:
                An LLM inference engine to infer risks from the use cases.
            new_prefix: str
                The CURIE prefix for the new list of risks
            mapping_method: MappingMethod
                The possible values for type of risk mapping method

        Returns:
            List[Mapping]
                Result containing a list of mappings
        """
        type_check(
            "<RAN28959363E>",
            InferenceEngine,
            allow_none=True,
            inference_engine=inference_engine,
        )
        value_check(
            "<RAN85167315E>",
            new_risks and existing_risks,
            "Please provide new_risks and existing_risks",
        )
        risk_mapper = RiskMapper(
            new_risks=new_risks,
            existing_risks=existing_risks,
            inference_engine=inference_engine,
            new_prefix=new_prefix,
            mapping_method=mapping_method,
        )

        return risk_mapper.generate(
            new_risks=new_risks,
            existing_risks=existing_risks,
            inference_engine=inference_engine,
            new_prefix=new_prefix,
            mapping_method=mapping_method,
        )

    def get_risk_incidents(cls, taxonomy: Optional[str] = None):
        """Get risk incident instances, optionally filtered by taxonomy

        Returns:
            List[RiskIncident]
                Result containing a list of AI Risk Incidents
        """
        type_check(
            "<RAN04811131E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        risk_incident_instances: List[RiskIncident] = (
            cls._risk_explorer.get_risk_incidents(taxonomy=taxonomy)
        )
        return risk_incident_instances

    def get_risk_incident(cls, id=None, taxonomy=None):
        """Get an risk incident instance filtered by risk incident id

        Args:
            id: str
                The string id identifying the risk incident
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            RiskIncident
                Result containing a risk incident.
        """
        type_check(
            "<RAN97353068E>",
            str,
            allow_none=False,
            id=id,
        )
        type_check(
            "<RAN38198685E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        risk_incident: RiskIncident | None = cls._risk_explorer.get_risk_incident(id=id)
        return risk_incident

    def get_related_risk_incidents(
        cls,
        risk=None,
        risk_id=None,
        taxonomy=None,
    ):
        """Get related risk incident filtered by risk id

        Args:
            risk: (Optional) Risk
                The risk
            risk_id: (Optional) str
                The string ID identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy
        Returns:
            List[RiskIncident]
                Result containing a list of AI risk incidents
        """
        type_check(
            "<RAN40791379E>",
            Risk,
            allow_none=True,
            risk=risk,
        )
        type_check(
            "<RANC9FDCC45E>",
            str,
            allow_none=True,
            risk_id=risk_id,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN79007538E>",
            risk or risk_id,
            risk or risk_id,
            "Please provide risk or id",
        )

        related_risk_incidents = cls._risk_explorer.get_related_risk_incidents(
            risk=risk,
            risk_id=risk_id,
            taxonomy=taxonomy,
        )
        return related_risk_incidents

    def get_all_evaluations(cls, taxonomy=None):
        """Get all evaluation definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[RiskControl]
                Result containing a list of AiEval
        """
        type_check("<RAN18094995E>", str, allow_none=True, taxonomy=taxonomy)

        evaluation_instances: list[AiEval] = cls._risk_explorer.get_all_evaluations(
            taxonomy
        )
        return evaluation_instances

    def get_evaluation(cls, id=None, taxonomy=None):
        """Get an evaluation definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the evaluation
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Action
                Result containing an evaluation.
        """
        type_check("<RAN84465757E>", str, allow_none=False, id=id)
        type_check("<RAN29906222E>", str, allow_none=True, taxonomy=taxonomy)

        evaluation: AiEval | None = cls._risk_explorer.get_evaluation(id=id)
        return evaluation

    def get_related_evaluations(cls, risk=None, risk_id=None, taxonomy=None):
        """Get related evaluations filtered by risk id

        Args:
            risk: (Optional) Risk
                The risk
            risk_id: (Optional) str
                The string ID identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy
        Returns:
            List[AiEval]
                Result containing a list of AI evaluations
        """
        type_check("<RAN04616807E>", Risk, allow_none=True, risk=risk)
        type_check(
            "<RAN05640166E>",
            str,
            allow_none=True,
            risk_id=risk_id,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN39630388E>",
            risk or risk_id,
            "Please provide risk or id",
        )

        related_evaluations = cls._risk_explorer.get_related_evaluations(
            risk=risk, risk_id=risk_id, taxonomy=taxonomy
        )
        return related_evaluations

    def get_benchmark_metadata_cards(cls, risk=None, risk_id=None, taxonomy=None):
        """Get all benchmark metadata definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[BenchmarkMetadataCard]
                Result containing a list of BenchmarkMetadataCards
        """
        type_check(
            "<RAN07894687E>",
            str,
            allow_none=True,
            risk_id=risk_id,
            taxonomy=taxonomy,
        )
        type_check("<RAN30190075E>", Risk, allow_none=True, risk=risk)

        benchmark_metatdata_card_instances: list[BenchmarkMetadataCard] = (
            cls._risk_explorer.get_all_benchmark_metadata_cards(taxonomy)
        )
        return benchmark_metatdata_card_instances

    def get_benchmark_metadata_card(cls, id=str):
        """Get an benchmark_metadata_card definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the benchmark_metadata_card
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            BenchmarkMetadataCard
                Result containing a benchmark_metadata_card.
        """
        type_check(
            "<RAN30946549E>",
            str,
            allow_none=False,
            id=id,
        )

        benchmark_metadata_card: BenchmarkMetadataCard | None = (
            cls._risk_explorer.get_benchmark_metadata_card(id=id)
        )
        return benchmark_metadata_card

    def get_documents(cls, taxonomy=None):
        """Get all document definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Documentation]
                Result containing a list of Documentation
        """
        type_check(
            "<RAN61770043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        document_instances: list[Documentation] = cls._risk_explorer.get_documents(
            taxonomy
        )
        return document_instances

    def get_document(cls, id=str):
        """Get a document definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the documentation entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Documentation
                Result containing a document.
        """
        type_check(
            "<RAN12472418E>",
            str,
            allow_none=False,
            id=id,
        )

        document: Documentation | None = cls._risk_explorer.get_document(id=id)
        return document

    def get_datasets(cls, taxonomy=None):
        """Get all dataset definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Dataset]
                Result containing a list of Dataset entries
        """
        type_check(
            "<RAN61770043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        dataset_instances: list[Dataset] = cls._risk_explorer.get_datasets(taxonomy)
        return dataset_instances

    def get_dataset(cls, id=str):
        """Get a dataset definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the dataset entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Dataset
                Result containing a dataset.
        """
        type_check(
            "<RAN12472418E>",
            str,
            allow_none=False,
            id=id,
        )

        dataset: Dataset | None = cls._risk_explorer.get_dataset(id=id)
        return dataset

    def get_stakeholders(cls, taxonomy=None):
        """Get all stakeholder definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Stakeholder]
                Result containing a list of Stakeholder entries
        """
        type_check(
            "<RAN61770043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        stakeholder_instances: list[Stakeholder] = cls._risk_explorer.get_stakeholders(
            taxonomy
        )
        return stakeholder_instances

    def get_stakeholder(cls, id=str):
        """Get a stakeholder definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the stakeholder entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Stakeholder
                Result containing a stakeholder.
        """
        type_check(
            "<RAN12472418E>",
            str,
            allow_none=False,
            id=id,
        )

        stakeholder: Stakeholder | None = cls._risk_explorer.get_stakeholder(id=id)
        return stakeholder

    def get_intrinsics(cls, taxonomy=None):
        """Get all intrinsic definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[LLMIntrinsic]
                Result containing a list of LLMIntrinsic entries
        """
        type_check(
            "<RAN61770043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        intrinsic_instances: list[LLMIntrinsic] = cls._risk_explorer.get_llmintrinsics(
            taxonomy
        )
        return intrinsic_instances

    def get_intrinsic(cls, id=str):
        """Get an intrinsic definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the intrinsic entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            LLMIntrinsic
                Result containing a intrinsic.
        """
        type_check(
            "<RAN12472418E>",
            str,
            allow_none=False,
            id=id,
        )

        intrinsic: LLMIntrinsic | None = cls._risk_explorer.get_llmintrinsic(id=id)
        return intrinsic

    def get_related_intrinsics(
        cls,
        risk=None,
        tag=None,
        risk_id=None,
        name=None,
        taxonomy=None,
    ):
        """Get related intrinsics for a risk definition from the LinkML.  The risk is identified by risk id, tag, or name

        Args:
            risk: (Optional) Risk
                The risk
            risk_id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            List
                Result containing a list of Intrinsics
        """
        type_check(
            "<RAN4E03158FE>",
            Risk,
            allow_none=True,
            risk=risk,
        )
        type_check(
            "<RAN55784808E>",
            str,
            allow_none=True,
            tag=tag,
            risk_id=risk_id,
            name=name,
            taxonomy=taxonomy,
        )
        value_check(
            "<RAN5DCADF94E>",
            risk or tag or risk_id or name,
            "Please provide risk, tag, risk_id, or name",
        )

        intrinsics = cls._risk_explorer.get_related_llmintrinsics(
            risk=risk,
            tag=tag,
            risk_id=risk_id,
            name=name,
            taxonomy=taxonomy,
        )
        return intrinsics

    def get_adapters(cls, taxonomy=None):
        """Get all adapter definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Adapter]
                Result containing a list of Adapter entries
        """
        type_check(
            "<RAN61770043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        adapter_instances: list[Adapter] = cls._risk_explorer.get_adapters(taxonomy)
        return adapter_instances

    def get_adapter(cls, id=str):
        """Get an adapter definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the stakeholder entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Adapter
                Result containing a adapter.
        """
        type_check(
            "<RAN12472418E>",
            str,
            allow_none=False,
            id=id,
        )

        adapter: Adapter | None = cls._risk_explorer.get_adapter(id=id)
        return adapter

    def get_llm_question_policies(cls, taxonomy=None):
        """Get all LLM Quesiton Policy definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[LLMQuestionPolicy]
                Result containing a list of LLMQuestionPolicy entries
        """
        type_check(
            "<RAN61796043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        llm_question_policy_instances: list[LLMQuestionPolicy] = (
            cls._risk_explorer.get_llm_question_policies(taxonomy)
        )
        return llm_question_policy_instances

    def get_llm_question_policy(cls, id=str):
        """Get an LLM Question Policy definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the LLM question policy entry
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            LLMQuestionPolicy
                Result containing an LLM Question Policy.
        """
        type_check(
            "<RAN32462418E>",
            str,
            allow_none=False,
            id=id,
        )

        llm_question_policy: LLMQuestionPolicy | None = (
            cls._risk_explorer.get_llm_question_policy(id=id)
        )
        return llm_question_policy

    def get_principles(cls, taxonomy=None, document=None):
        """Get all Principle definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy
            document: str
                (Optional) The string label for a document


        Returns:
            list[Principle]
                Result containing a list of Principle entries
        """
        type_check(
            "<RAN61573043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )
        type_check(
            "<RAN62577043E>",
            str,
            allow_none=True,
            document=document,
        )

        principle_instances: list[Principle] = cls._risk_explorer.get_principles(
            taxonomy, document
        )
        return principle_instances

    def get_principle(cls, id=str):
        """Get a Principle definition from the LinkML, filtered by id

        Args:
            id: str
                The string id identifying the Principle entry

        Returns:
            Principle
                Result containing a Principle.
        """
        type_check(
            "<RAN97462678E>",
            str,
            allow_none=False,
            id=id,
        )

        principle: Principle | None = cls._risk_explorer.get_principle(id=id)
        return principle

    def get_instances(cls, target_class, taxonomy=None):
        """Get all instance definitions from the LinkML

        Args:
            target_class: str
                (Optional) The string label for a target class
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Any]
                Result containing a list of instance entries
        """
        type_check(
            "<RAN92358069E>",
            str,
            allow_none=False,
            target_class=id,
        )
        type_check(
            "<RAN61877043E>",
            str,
            allow_none=True,
            taxonomy=taxonomy,
        )

        instances: list[Any] = cls._risk_explorer.get_instances(target_class, taxonomy)
        return instances

    def identify_domain_from_usecases(
        cls, usecases: List[str], inference_engine: InferenceEngine, verbose=True
    ) -> List[List[str]]:
        """Identify potential risks from a usecase description

        Args:
            usecases (List[str]):
                A List of strings describing AI usecases
            inference_engine (InferenceEngine):
                An LLM inference engine to identify AI tasks from usecases.

        Returns:
            List[List[str]]:
                Result containing a list of AI tasks
        """
        type_check(
            "<RAN3B9CD886E>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check(
            "<RAN4CDA6852E>",
            List,
            allow_none=False,
            usecases=usecases,
        )
        value_check(
            "<RAN0E435F50E>",
            inference_engine and usecases,
            "Please provide usecases and inference_engine",
        )

        # Load risk questionnaire CoT from the template dir
        risk_questionnaire = load_resource("risk_questionnaire_cot.json")

        # Retrieve domain question data
        domain_ques_data = risk_questionnaire[0]

        # Prepare few shots inference prompts from CoT Data
        prompts = [
            FewShotPromptBuilder(
                prompt_template=QUESTIONNAIRE_COT_TEMPLATE,
            ).build(
                cot_examples=domain_ques_data["cot_examples"],
                usecase=usecase,
                question=domain_ques_data["question"],
            )
            for usecase in usecases
        ]

        # Invoke inference service
        return inference_engine.chat(
            messages=prompts,
            response_format=DOMAIN_TYPE_SCHEMA,
            postprocessors=["json_object"],
            verbose=verbose,
        )

    def categorize_risk_severity(
        self,
        usecases: List[str],
        inference_engine: InferenceEngine,
    ):
        """Determine the severity of risks based on the use case description.
        Args:
            usecases (List[str]):
                A List of strings describing AI usecases
            inference_engine (InferenceEngine):
                An LLM inference engine
        Returns:
            results (List[Dict]):
                Results detailing risk categorization by usecase.
        """
        type_check(
            "<RAN75727859E>",
            InferenceEngine,
            allow_none=False,
            inference_engine=inference_engine,
        )
        type_check(
            "<RAN68734549E>",
            List,
            allow_none=False,
            usecases=usecases,
        )
        value_check(
            "<RAN30508300E>",
            inference_engine and usecases,
            "Please provide usecases and inference_engine",
        )

        # Create Risk Severity instance
        risk_severity = RiskSeverityCategorizer(inference_engine)

        # Load risk questionnaire from the template dir
        risk_questionnaire = load_resource("risk_questionnaire_cot.json")

        # Collecting required parameters for categorizing risk severity per usecase
        results = []
        for usecase in usecases:

            # Get AI Domain of the usecase
            domain_predictions = [
                domain.prediction["answer"]
                for domain in self.identify_domain_from_usecases(
                    [usecase], inference_engine=inference_engine, verbose=False
                )
            ]
            domain = domain_predictions[0] if len(domain_predictions) == 1 else None

            # Using a risk questionnaire to identify key attributes necessary for
            # constituting an AI system from the usecase.
            #
            # (Q4) AI User that interacts with the AI system
            # (Q5) Intended purpose of the AI System
            # (Q6) The capability of an AI system to do what it is designed to do
            # (Q7) AI Subject impacted by the AI System
            predictions = [
                response.prediction["answer"]
                for response in self.generate_few_shot_risk_questionnaire_output(
                    usecase,
                    list(
                        filter(
                            lambda question: question["no"] in ["Q4", "Q5", "Q6", "Q7"],
                            risk_questionnaire,
                        )
                    ),
                    inference_engine=inference_engine,
                    verbose=False,
                )
            ]

            # Extracting predictions
            if len(predictions) == 4:
                ai_user = predictions[0]
                purpose = predictions[1]
                capability = predictions[2]
                ai_subject = predictions[3]
            else:
                raise Exception(
                    "Unable to retrieve all the required attributes from the usecase. Please try again."
                )

            # Calling the risk categorization API with the required attributes.
            results.append(
                risk_severity.categorize(
                    domain, purpose, capability, ai_user, ai_subject
                )
            )

        return results

    def run_ares_evaluation(
        cls, risks: List[Risk], inference_engine: InferenceEngine, target
    ) -> None:
        """Submit potential attack risks for ARES red-teaming evaluation.
        This API needs the `ran-ares-integration` extension. Please install this extension via
        `ran-extension install ran-ares-integration`. Refer `Readme` file for further information
        on `ai-atlas-nexus` extensions.

        Args:
            risks (List[Risk]):
                A List of attack risks
            inference_engine (InferenceEngine):
                An instance of the LLM inference engine
            target (Optional[Path], optional):
                A target AI model to perform the ARES red-teaming evaluation

        Returns:
            None
        """
        logger.info(
            f"Risks submitted for ARES evluation: {json.dumps([risk.name for risk in risks], indent=2)}"
        )

        # Load RAN-ARES extension
        ares_extension = Extension.load(
            "ran-ares-integration", inference_engine, target=target
        )

        # Run extension on the given risks
        for risk in risks:
            ares_extension.run(risk)



# Standard Library
import csv

# Third Party
from linkml_runtime.dumpers import YAMLDumper
from linkml_runtime.loaders import yaml_loader

# Local
from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import *


SCHEMA_DIR = "src/ai_risk_ontology/schema/"
DATA_DIR = "src/data/knowledge-graph/"


def create_container_object() -> Container:

    # Create risk-related actions
    action_list = get_actions()
    risks = get_risks()
    actions = [
        Action(
            id=action["NIST Action ID"],
            name=action["NIST Action ID"],
            description=action["NIST Title"],
            hasAiActorTask=[
                task.strip() for task in action["AI Actor Tasks"].split(",")
            ],
            hasRelatedRisk=get_related_risks(action["NIST GAI Risks"], risks),
            hasDocumentation=["NIST.AI.600-1"],
        )
        for action in action_list
    ]
    # Create container
    container = Container(actions=actions)
    return container


def get_risks() -> dict[str, str]:
    instances = yaml_loader.load(DATA_DIR + "nist_ai_rmf_data.yaml", Container)
    return {risk.name: risk.id for risk in instances.risks}


def get_related_risks(related_risks_str: str, risks: dict[str, str]) -> list[str]:
    related_risks = related_risks_str.split(";")
    return [
        id
        for name, id in risks.items()
        for risk in related_risks
        if name == risk.strip()
    ]


def get_actions() -> list[dict]:
    with open("resources/actions_extracted_from_nist.csv") as csvfile:
        importer = csv.DictReader(csvfile)
        return [row for row in importer]


with open(
    DATA_DIR + "nist_ai_rmf_actions_data.yaml", "+tw", encoding="utf-8"
) as output_file:
    container = create_container_object()
    print(YAMLDumper().dumps(container), file=output_file)
    output_file.close()



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import pydantic
import requests

from ai_atlas_nexus.blocks.inference.params import (
    InferenceEngineCredentials,
    OllamaInferenceEngineParams,
    OpenAIChatCompletionMessageParam,
    RITSInferenceEngineParams,
    TextGenerationInferenceOutput,
    VLLMInferenceEngineParams,
    WMLInferenceEngineParams,
)
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)


class InferenceEngine(ABC):

    def __init__(
        self,
        model_name_or_path: str,
        credentials: Optional[Union[Dict, InferenceEngineCredentials]] = None,
        parameters: Optional[
            Union[
                RITSInferenceEngineParams,
                WMLInferenceEngineParams,
                OllamaInferenceEngineParams,
                VLLMInferenceEngineParams,
            ]
        ] = None,
        think: Optional[bool] = None,
        concurrency_limit: int = 10,
    ):
        """Create an instance of the InferenceEngine using the `model_name_or_path` and chosen LLM service.

        Args:
            model_name_or_path (str): model name or path as per the LLM model service
            credentials (Optional[Union[Dict, InferenceEngineCredentials]], optional): credentials for the inference engine instance. Defaults to None.
            parameters (Optional[ Union[ RITSInferenceEngineParams, WMLInferenceEngineParams, OllamaInferenceEngineParams, VLLMInferenceEngineParams, ] ], optional): parameters to use during request generation. Defaults to None.
            think (Optional[bool], optional): enable or disable model thinking. Currently, only supported in Ollama. Defaults to None.
            concurrency_limit (int, optional): No of parallel calls to be made to the LLM service. Defaults to 10.
        """

        self.model_name_or_path = model_name_or_path
        self.credentials = self.prepare_credentials(credentials or {})
        self.parameters = self._check_if_parameters_are_valid(parameters or {})
        self.think = think
        self.concurrency_limit = concurrency_limit

        # Create inference client
        self.client = self.create_client(self.credentials)

        # Health check
        try:
            self.ping()
        except Exception as e:
            raise Exception(
                f"Failed to create `{self.__class__.__name__}`. Reason: {str(e)} Given API credentials: {self.credentials}"
            )

        # Verify whether the inference engine and the model type support `thinking`.
        if think:
            self.is_thinking_supported()

        logger.info(f"Created {self._inference_engine_type} inference engine.")

    def _check_if_parameters_are_valid(self, parameters):
        if parameters:
            invalid_params = []
            for param_key, _ in parameters.items():
                if param_key not in list(
                    self._inference_engine_parameter_class.__annotations__
                ):
                    invalid_params.append(param_key)

            if len(invalid_params) > 0:
                raise Exception(
                    f"Invalid parameters found: {invalid_params}. {self._inference_engine_type} inference engine only supports {list(self._inference_engine_parameter_class.__annotations__)}"
                )

        return parameters

    def _to_openai_format(self, prompt: Union[OpenAIChatCompletionMessageParam, str]):
        if isinstance(prompt, str):
            return [{"role": "user", "content": prompt}]
        elif pydantic.TypeAdapter(OpenAIChatCompletionMessageParam).validate_python(
            prompt
        ):
            return prompt
        else:
            raise Exception(
                f"Invalid input format: {prompt}. Please use openai format or plain str."
            )

    def ping(self):
        # Implement inference engine specific ping in their respective class.
        pass

    def is_thinking_supported(self):
        raise Exception(
            f"Currently, model thinking (think=True) is only supported in OllamaInferenceEngine."
        )

    @abstractmethod
    def prepare_credentials(
        self,
        credentials: Union[Dict, InferenceEngineCredentials],
    ) -> InferenceEngineCredentials:
        raise NotImplementedError

    @abstractmethod
    def create_client(self, credentials: InferenceEngineCredentials) -> Any:
        raise NotImplementedError

    @abstractmethod
    def generate(
        self,
        prompts: List[str],
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> List[TextGenerationInferenceOutput]:
        raise NotImplementedError

    @abstractmethod
    def chat(
        self,
        messages: Union[
            List[OpenAIChatCompletionMessageParam],
            List[str],
        ],
        tools=None,
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> List[TextGenerationInferenceOutput]:
        raise NotImplementedError



from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import EuAiRiskCategory


LIST_OF_STR_SCHEMA = {
    "type": "array",
    "items": {"enum": None},
}

QUESTIONNAIRE_OUTPUT_SCHEMA = {
    "type": "object",
    "properties": {
        "answer": {"type": "string"},
        "explanation": {"type": "string"},
        "confidence": {"type": "string"},
    },
    "required": [
        "answer",
        "explanation",
        "confidence"
    ],
}

DOMAIN_TYPE_SCHEMA = {
    "type": "object",
    "properties": {
        "answer": {
            "type": "string",
            "enum": [
                "Customer service/support",
                "Technical",
                "Information retrieval",
                "Strategy",
                "Code/software engineering",
                "Communications",
                "IT/business automation",
                "Writing assistant",
                "Financial",
                "Talent and Organization including HR",
                "Product",
                "Marketing",
                "Cybersecurity",
                "Healthcare",
                "User Research",
                "Sales",
                "Risk and Compliance",
                "Design",
                "Other",
            ],
        },
        "explanation": {"type": "string"},
    },
    "required": [
        "answer",
        "explanation",
    ],
}

RISK_CATEGORY_SCHEMA = {
    "type": "object",
    "properties": {
        "Description": {"type": "string"},
        "Classification": {
            "type": "string",
            "enum": [
                EuAiRiskCategory.EXCLUDED.value,
                EuAiRiskCategory.PROHIBITED.value,
                EuAiRiskCategory.HIGH_RISK_EXCEPTION.value,
                EuAiRiskCategory.HIGH_RISK.value,
                EuAiRiskCategory.LIMITED_OR_LOW_RISK.value,
            ],
        },
        "AIActText": {"type": "string"},
        "Reasoning": {"type": "string"},
    },
    "required": [
        "Description",
        "Classification",
        "AIActText",
        "Reasoning",
    ],
}



import glob
import os

from linkml_runtime.loaders import yaml_loader

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Container
from ai_atlas_nexus.data import get_data_path
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)


def load_yamls_to_container(base_dir):
    """Function to load the AIAtlasNexus with data

    Args:
        base_dir: str
            (Optional) user defined base directory path

    Returns:
        YAMLRoot instance of the Container class
    """

    # Get system yaml data path
    system_data_path = get_data_path()

    master_yaml_files = []
    for yaml_dir in [system_data_path, base_dir]:
        # Include YAML files from the user defined `base_dir` if exist.
        if yaml_dir is not None:
            master_yaml_files.extend(
                glob.glob(os.path.join(yaml_dir, "**", "*.yaml"), recursive=True)
            )

    yml_items_result = {}
    for yaml_file in master_yaml_files:
        try:
            yml_items = yaml_loader.load_as_dict(source=yaml_file)
            for ontology_class, instances in yml_items.items():
                yml_items_result.setdefault(ontology_class, []).extend(instances)
        except Exception as e:
            logger.info(f"YAML ignored: {yaml_file}. Failed to load. {e}")

    # TODO: generalise this to cover all ontology classes

    # combine any risk entries which share the same id, for example a risk, and a secondary entry for a mapping
    combine_risks = {}

    for risk in yml_items_result["risks"]:
        risk_id = risk["id"]

        if risk_id not in combine_risks:
            combine_risks[risk_id] = {"id": risk_id}

        for key, value in risk.items():
            if key != "id":
                if key not in combine_risks[risk_id]:
                    combine_risks[risk_id][key] = value
                else:
                    if combine_risks[risk_id][key] is not None:
                        combine_risks[risk_id][key] = [
                            *combine_risks[risk_id][key],
                            *value,
                        ]
                    else:
                        combine_risks[risk_id][key] = value

    yml_items_result["risks"] = list(combine_risks.values())

    ontology = yaml_loader.load_any(
        source=yml_items_result,
        target_class=Container,
    )

    return ontology



import importlib
import subprocess
import sys

from typer import Typer

from ai_atlas_nexus.toolkit.logging import configure_logger


app = Typer()
logger = configure_logger(__name__)


@app.callback()
def main() -> None:
    """
    RAN Extension CLI Application
    """


@app.command()
def install(extension_name: str) -> None:
    """
    Installs an extension

    :param extension_name: Name of the extension to install
    """
    logger.info("Installing extension: %s", extension_name)

    extension_location = (
        "git+ssh://git@github.com/IBM/ai-atlas-nexus-extensions.git"
        + "#subdirectory="
        + extension_name
    )
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", extension_location]
    )  # nosec


class Extension:
    """
    Imports and loads an AI atlas nexus extension main class
    """

    @staticmethod
    def load(extension_name: str, *args, **kwargs) -> type:
        """Import the main extension class.

        Args:
            extension_name (str): an AI Atlas Nexus extension name

        Raises:
            ModuleNotFoundError: if the main module is not present
            TypeError: if the main Extension class is not present

        Returns:
            type: The class for instantiation.
        """

        try:
            # Looking for an absolute class path
            module = importlib.import_module(extension_name.replace("-", "_") + ".main")
        except ModuleNotFoundError as no_mod:
            logger.error(f"Following extension not found: {extension_name}")
            logger.error(f"Install with: ran-extension install {extension_name}")
            sys.exit()

        try:
            extension_class = getattr(module, "Extension")
        except:
            raise TypeError(
                f"Error loading extension class from the main module: '{extension_name.replace('-', '_')}.main.Extension'. Please contact the extension owner."
            )

        return extension_class(*args, **kwargs)



import logging

import logzero


def configure_logger(
    logger_name: str,
    logging_level="INFO",
    json=False,
):
    if logger_name is None or logger_name == "":
        raise Exception(
            "Logger name cannot be None or empty. Accessing root logger is restricted. Please use routine configure_root_logger() to access the root logger."
        )

    logging_level = logging.getLevelNamesMapping()[logging_level.upper()]

    log_format = (
        "%(color)s[%(asctime)s:%(msecs)d] - %(levelname)s - AIAtlasNexus - "
        "%(end_color)s%(message)s"
    )

    formatter = logzero.LogFormatter(fmt=log_format, datefmt="%Y-%m-%d %H:%M:%S")
    return logzero.setup_logger(
        name=logger_name, level=logging_level, formatter=formatter, json=json
    )



"""
Generate Cypher code from linkml instance specifications.
"""

# Standard Library
from os import listdir, makedirs
from os.path import isfile, join
from pathlib import Path
from typing import Any

# Third Party
from cymple import QueryBuilder
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.utils.schemaview import SchemaView
from pydantic import BaseModel

# Local
from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Container
from ai_atlas_nexus.toolkit.data_utils import load_yamls_to_container
from ai_atlas_nexus.toolkit.logging import configure_logger


MAPPING_DIR = "src/ai_atlas_nexus/data/mappings/"
OUTPUT_DIR = "graph_export/cypher/"
SCHEMA_DIR = "src/ai_atlas_nexus/ai_risk_ontology/schema/"
SCHEMA_FILE = "ai-risk-ontology.yaml"
_log = configure_logger("ExportCypher")


class GraphEdge:
    label: str
    source_id: str
    source_label: str
    target_id: str
    target_label: str

    def __init__(
        self,
        label: str,
        source_id: str,
        source_label: str,
        target_id: str,
        target_label: str,
    ):
        self.label = label
        self.source_id = source_id
        self.source_label = source_label
        self.target_id = target_id
        self.target_label = target_label

    def __str__(self):
        return f"{self.label}: {self.source_id}/{self.source_label} -> {self.target_id}/{self.target_label}"

    def to_cypher(self) -> str:
        return f'MATCH (src: {self.source_label} {{id: "{self.source_id}"}}) MATCH (dst: {self.target_label} {{id: "{self.target_id}"}}) MERGE (src)-[: {self.label}]->(dst);\n'


class GraphNode:
    label: str
    id: str
    properties: dict
    edges: list[GraphEdge]

    def __init__(
        self,
        entity_id: str,
        label: str,
        properties: dict[str, Any],
        relations: list[GraphEdge],
    ):
        self.label = label
        self.id = entity_id
        self.properties = properties
        self.edges = relations

    def to_cypher(self, with_relations: bool = True) -> str:
        merge_node = "MERGE (node:" + self.label + ' {id: "' + self.id + '"})'
        if self.properties:
            merge_node += (
                " ON CREATE SET node += {"
                + ",".join(
                    f'{key}: "{value}"' for key, value in self.properties.items()
                )
                + "}"
            )
        merge_edges = (
            ";\n" + "".join(edge.to_cypher() for edge in self.edges)
            if with_relations
            else ""
        )
        return merge_node + merge_edges

    def __eq__(self, value):
        if isinstance(value, GraphNode):
            return (value.label == self.label) and (value.id == self.id)
        return False

    def __hash__(self):
        return (self.label + "_" + self.id).__hash__()


def get_linkml_types(schema_view: SchemaView) -> list[str]:
    type_names = list(node.name for node in schema_view.all_types().values())
    type_names.extend(str(node.uri) for node in schema_view.all_types().values())
    return list(set(type_names))


def is_relationship(
    schema_view: SchemaView,
    linkml_class: str,
    linkml_slot: str,
    linkml_types: list[str],
) -> bool:
    class_slots = schema_view.class_induced_slots(linkml_class)
    assert linkml_slot in [
        slot.name for slot in class_slots
    ], f"{linkml_slot} is not a known slot for class {linkml_class}."
    slot_def = [slot for slot in class_slots if slot.name == linkml_slot].pop()
    if slot_def.range not in linkml_types:
        return True
    return False


def convert_entity_to_graph_node(
    entity: BaseModel, label: str, schema_view: SchemaView, linkml_types: list[str]
) -> list[GraphNode]:
    return_list: list[GraphNode] = []
    # Extract properties, namely slots that have a generic LinkML type as range
    properties = {
        item: entity.__getattribute__(item)
        for item in entity.model_dump(exclude={"id"}).keys()
        if not is_relationship(schema_view, label, item, linkml_types)
        and entity.__getattribute__(item) is not None
    }

    # Extract relationships, namely slots that don't have a generic LinkML type as range
    relations = []
    relation_slots = [
        item
        for item in entity.model_dump(exclude={"id"}).keys()
        if is_relationship(schema_view, label, item, linkml_types)
        and entity.__getattribute__(item) is not None
    ]
    # Retrieve all LinkML slot definitions
    slot_definitions = schema_view.class_induced_slots(label)
    for slot_name in relation_slots:
        slot_definition = [
            item for item in slot_definitions if item.name == slot_name
        ].pop()
        if (
            slot_definition.range == "Any"
            and slot_definition.slot_uri is not None
            and slot_definition.slot_uri.startswith("skos:")
        ):
            dst_label = str(slot_definition.owner)
        else:
            dst_label = str(slot_definition.range)
        if slot_definition.multivalued:
            # slot denotes multiple relationships
            if slot_definition.inlined_as_list:
                # Target node is defined here
                return_list.extend(
                    [
                        GraphNode(
                            item.id,
                            dst_label,
                            {
                                prop_name: item.__getattribute__(prop_name)
                                for prop_name in item.model_dump(exclude={"id"}).keys()
                                if not is_relationship(
                                    schema_view,
                                    dst_label,
                                    prop_name,
                                    linkml_types,
                                )
                                and item.__getattribute__(prop_name) is not None
                            },
                            [],
                        )
                        for item in entity.__getattribute__(slot_name)
                    ]
                )
                relations.extend(
                    [
                        GraphEdge(
                            slot_name,
                            entity.__getattribute__("id"),
                            label,
                            item.id,
                            dst_label,
                        )
                        for item in entity.__getattribute__(slot_name)
                    ]
                )
            else:
                relations.extend(
                    [
                        GraphEdge(
                            slot_name,
                            entity.__getattribute__("id"),
                            label,
                            item,
                            dst_label,
                        )
                        for item in entity.__getattribute__(slot_name)
                    ]
                )
        else:
            relations.append(
                GraphEdge(
                    slot_name,
                    entity.__getattribute__("id"),
                    label,
                    entity.__getattribute__(slot_name),
                    dst_label,
                )
            )
    return_list.append(
        GraphNode(entity.__getattribute__("id"), label, properties, relations)
    )
    return return_list


def export_data_to_cypher(container: Container) -> str:
    file_list = [
        file_name
        for file_name in listdir(SCHEMA_DIR)
        if isfile(join(SCHEMA_DIR, file_name))
    ]
    importmap = {Path(item).stem: SCHEMA_DIR + Path(item).stem for item in file_list}
    model: SchemaDefinition = yaml_loader.load(
        SCHEMA_FILE, SchemaDefinition, base_dir=SCHEMA_DIR
    )
    schema_view = SchemaView(schema=model, merge_imports=True, importmap=importmap)
    linkml_types = get_linkml_types(schema_view)
    container_class = model.classes["Container"]
    graph_nodes: list[GraphNode] = []
    for container_slot in schema_view.class_induced_slots("Container"):
        if container_slot.range not in linkml_types:
            print("Processing entities of type " + container_slot.name)
            entities = container.__getattribute__(container_slot.name)
            if not entities:
                continue
            graph_nodes.extend(
                [
                    node
                    for item in entities
                    for node in convert_entity_to_graph_node(
                        item, str(container_slot.range), schema_view, linkml_types
                    )
                ]
            )

    # First pass: Generate Cypher code for creation of the nodes
    cypher_code = ";\n".join(
        [graph_node.to_cypher(with_relations=False) for graph_node in graph_nodes]
    )
    if cypher_code:
        cypher_code += ";\n"

    # Second pass: Generate Cypher code for creation of relations
    for graph_node in graph_nodes:
        if graph_node.edges:
            cypher_code += "".join([edge.to_cypher() for edge in graph_node.edges])
    return cypher_code


if __name__ == "__main__":
    ontology = load_yamls_to_container(MAPPING_DIR)
    makedirs(OUTPUT_DIR, exist_ok=True)
    with open(
        OUTPUT_DIR + "ai-risk-ontology.cypher", "+tw", encoding="utf-8"
    ) as output_file:
        print(export_data_to_cypher(ontology), file=output_file)
        output_file.close()



# Standard Library
from os import listdir
from os.path import isfile, join
from pathlib import Path

# Third Party
from linkml_runtime.dumpers import YAMLDumper
from pydantic import BaseModel
from sssom.parsers import parse_sssom_table

# Local
from ai_atlas_nexus.ai_risk_ontology import Container, Risk
from ai_atlas_nexus.toolkit.logging import configure_logger


MAP_DIR = "src/ai_atlas_nexus/data/mappings/"
DATA_DIR = "src/ai_atlas_nexus/data/knowledge_graph/mappings/"

logger = configure_logger(__name__)


class RiskMap(BaseModel):
    src_risk_id: str
    target_risk_id: str
    relationship: str

    def __init__(self, src_risk_id: str, target_risk_id: str, relationship: str):
        src_id = src_risk_id.split(":")[-1]
        target_id = target_risk_id.split(":")[-1]

        super().__init__(
            src_risk_id=src_id,
            target_risk_id=target_id,
            relationship=relationship,
        )


def process_mapping_from_tsv_to_risk_mapping(file_name):
    tsv_file_name = join(MAP_DIR, file_name)
    mapping_set_df = parse_sssom_table(file_path=tsv_file_name)
    ms = mapping_set_df.to_mapping_set()
    risk_maps = [
        RiskMap(
            **{
                "src_risk_id": item["subject_id"],
                "target_risk_id": item["object_id"],
                "relationship": item["predicate_id"],
            }
        )
        for item in ms.mappings
        if item["predicate_id"] != "noMatch"
    ]
    return risk_maps


def process_mappings_to_risks(risk_maps):
    output_risks = []
    for rm in risk_maps:

        s_id = rm.src_risk_id
        o_id = rm.target_risk_id
        risk = Risk(id=s_id)
        risk_for_inverse = Risk(id=o_id)

        relationship = rm.relationship
        if relationship == "skos:closeMatch":
            risk.closeMatch = [o_id]
            risk_for_inverse.closeMatch = [s_id]
        elif relationship == "skos:exactMatch":
            risk.exactMatch = [o_id]
            risk_for_inverse.exactMatch = [s_id]
        elif relationship == "skos:broadMatch":
            risk.broadMatch = [o_id]
            risk_for_inverse.narrowMatch = [s_id]
        elif relationship == "skos:narrowMatch":
            risk.narrowMatch = [o_id]
            risk_for_inverse.broadMatch = [s_id]
        elif relationship == "skos:relatedMatch":
            risk.relatedMatch = [o_id]
            risk_for_inverse.relatedMatch = [s_id]
        else:
            logger.info("Unparseable predicate_id: %s", relationship)

        output_risks.append(risk)
        output_risks.append(risk_for_inverse)

    return output_risks


def write_to_file(output_risks, output_file):
    with open(output_file, "+tw", encoding="utf-8") as output_file:
        print(YAMLDumper().dumps(Container(risks=output_risks)), file=output_file)
        output_file.close()


if __name__ == "__main__":
    logger.info(f"Processing mapping files in : %s", MAP_DIR)
    mapping_files = [
        file_name
        for file_name in listdir(MAP_DIR)
        if (file_name.endswith(".md") == False) and isfile(join(MAP_DIR, file_name))
    ]
    for file_name in mapping_files:
        output_file = DATA_DIR + Path(file_name).stem + "_from_tsv_data.yaml"
        rs = process_mapping_from_tsv_to_risk_mapping(file_name)
        logger.info(f"Processed file: %s, %s valid entries", file_name, len(rs))
        output_risks = process_mappings_to_risks(rs)
        write_to_file(output_risks, output_file)



from abc import ABC, abstractmethod

from sssom_schema import Mapping

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Risk
from ai_atlas_nexus.blocks.inference.base import InferenceEngine
from ai_atlas_nexus.metadata_base import MappingMethod
from ai_atlas_nexus.toolkit.logging import configure_logger


logger = configure_logger(__name__)


class RiskMappingBase(ABC):

    def __init__(
        self,
        new_risks: list[Risk],
        existing_risks: list[Risk],
        inference_engine: InferenceEngine,
        new_prefix: str,
        mapping_method: MappingMethod,
    ):
        self.inference_engine = inference_engine
        self._new_risks = new_risks
        self._existing_risks = existing_risks
        self._new_prefix = new_prefix
        self._mapping_method = mapping_method

    @abstractmethod
    def generate(
        self,
        new_risks: list[Risk],
        existing_risks: list[Risk],
        inference_engine: InferenceEngine,
        new_prefix: str,
        mapping_method: MappingMethod,
    ) -> list[Mapping]:
        raise NotImplementedError



from .base import InferenceEngine
from .ollama import OllamaInferenceEngine
from .params import TextGenerationInferenceOutput
from .rits import RITSInferenceEngine
from .vllm import VLLMInferenceEngine
from .wml import WMLInferenceEngine



import json
from typing import List

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Risk
from ai_atlas_nexus.blocks.inference import TextGenerationInferenceOutput
from ai_atlas_nexus.blocks.prompt_response_schema import LIST_OF_STR_SCHEMA
from ai_atlas_nexus.blocks.prompt_templates import RISK_IDENTIFICATION_TEMPLATE
from ai_atlas_nexus.blocks.risk_detector import RiskDetector


class GenericRiskDetector(RiskDetector):

    def detect(self, usecases: List[str]) -> List[List[Risk]]:
        prompts = [
            self.prompt_builder(prompt_template=RISK_IDENTIFICATION_TEMPLATE).build(
                cot_examples=self._examples,
                usecase=usecase,
                risks=json.dumps(
                    [
                        {"category": risk.name, "description": risk.description}
                        for risk in self._risks
                    ],
                    indent=4,
                ),
                max_risk=self.max_risk,
            )
            for usecase in usecases
        ]

        # Populate schema items
        json_schema = dict(LIST_OF_STR_SCHEMA)
        json_schema["items"]["enum"] = [risk.name for risk in self._risks]

        # Invoke inference service
        inference_responses: List[TextGenerationInferenceOutput] = (
            self.inference_engine.generate(
                prompts,
                response_format=json_schema,
                postprocessors=["list_of_str"],
            )
        )

        return [
            (
                list(
                    filter(
                        lambda risk: risk.name in inference_response.prediction,
                        self._risks,
                    )
                )
                if inference_response.prediction
                else []
            )
            for inference_response in inference_responses
        ]



# Third Party
from linkml_runtime.dumpers import YAMLDumper
from linkml_runtime.loaders import yaml_loader

# Local
from ai_atlas_nexus.ai_risk_ontology import Action, Container, Risk


SCHEMA_DIR = "src/ai_risk_ontology/schema/"
DATA_DIR = "src/data/knowledge-graph/"


def create_container_object() -> Container:

    # Add risk-related actions to risks
    actions = get_actions()
    risk_data = get_risks()
    new_risks = []
    for risk in risk_data.risks:
        risk.hasRelatedAction = [
            action.id
            for action in actions
            if action.hasRelatedRisk is not None and risk.id in action.hasRelatedRisk
        ]
        new_risks.append(risk)
    # Create container
    container = Container(
        risks=new_risks, documents=risk_data.documents, taxonomies=risk_data.taxonomies
    )
    return container


def get_risks() -> Container:
    instances = yaml_loader.load(DATA_DIR + "nist_ai_rmf_data.yaml", Container)
    return instances


def get_related_risks(related_risks_str: str, risks: dict[str, str]) -> list[str]:
    related_risks = related_risks_str.split(";")
    return [
        id
        for name, id in risks.items()
        for risk in related_risks
        if name == risk.strip()
    ]


def get_actions() -> list[Action]:
    instances = yaml_loader.load(DATA_DIR + "nist_ai_rmf_actions_data.yaml", Container)
    return instances.actions


container = create_container_object()
with open(DATA_DIR + "nist_ai_rmf_data.yaml", "+tw", encoding="utf-8") as output_file:
    print(YAMLDumper().dumps(container), file=output_file)
    output_file.close()



import json
import os
from pathlib import Path


PACKAGEDIR = Path(__file__).parent.absolute()


def get_data_path():
    return os.path.join(PACKAGEDIR, "knowledge_graph")


def load_resource(file_name):
    return json.loads(
        Path(os.path.join(PACKAGEDIR, "templates", file_name)).read_text()
    )



# Standard Library
import os
import re

# Third Party
import pandas as pd
from datasets import load_dataset
from linkml_runtime.dumpers import YAMLDumper

# Local
from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import *


SCHEMA_DIR = "src/ai_risk_ontology/schema/"
DATA_DIR = "src/ai_atlas_nexus/data/knowledge-graph/"
ds = load_dataset("stanford-crfm/air-bench-2024", "judge_prompts", split="test")


def create_container_object() -> Container:
    # Create risk taxonomy paper
    documents = [
        Documentation(
            **{
                "id": "arxiv.org/pdf/2406.17864",
                "name": "The AI Risk Taxonomy (AIR 2024)",
                "description": "We present a comprehensive AI risk taxonomy derived from eight government policies from the European Union, United States, and China and 16 company policies worldwide, making a significant step towards establishing a unified language for generative AI safety evaluation. We identify 314 unique risk categories, organized into a four-tiered taxonomy. At the highest level, this taxonomy encompasses System & Operational Risks, Content Safety Risks, Societal Risks, and Legal & Rights Risks. The taxonomy establishes connections between various descriptions and approaches to risk, highlighting the overlaps and discrepancies between public and private sector conceptions of risk. By providing this unified framework, we aim to advance AI safety through information sharing across sectors and the promotion of best practices in risk mitigation for generative AI models and systems.",
                "url": "https://arxiv.org/pdf/2406.17864",
                "dateCreated": "2024-09-05",
                "dateModified": "2024-09-05",
            }
        )
    ]

    # Create risk taxonomy
    taxonomies = [
        RiskTaxonomy(
            **{
                "id": "ai-risk-taxonomy",
                "name": "The AI Risk Taxonomy (AIR 2024)",
                "description": "An AI risk taxonomy derived from eight government policies from the European Union, United States, and China and 16 company policies worldwide. It identifies 314 unique risk categories organized into a four-tiered taxonomy. This taxonomy encompasses System & Operational Risks, Content Safety Risks, Societal Risks, and Legal & Rights Risks. The taxonomy establishes connections between various descriptions and approaches to risk, highlighting the overlaps and discrepancies between public and private sector conceptions of risk. ",
                "url": "https://arxiv.org/pdf/2406.17864",
                "version": "1",
                "dateCreated": "2024-09-05",
                "hasDocumentation": ["arxiv.org/pdf/2406.17864"],
            }
        )
    ]

    # Create risk groups
    risk_groups = []
    risk_objects = []

    risks = get_risks()
    df = pd.DataFrame(
        columns=["cate-idx", "l2-name", "l3-name", "l4-name", "definition"]
    )
    df = pd.concat([df, pd.DataFrame(risks)], ignore_index=True)

    l2_grouping = df.groupby("l2-name", as_index=False)[["l3-name"]].agg(list)

    for i, group_l2 in l2_grouping.iterrows():
        risk_groups.append(
            RiskGroup(
                **{
                    "id": "ai-risk-taxonomy-"
                    + group_l2["l2-name"].lower().replace(" ", "-"),
                    "name": group_l2["l2-name"],
                    "isDefinedByTaxonomy": "ai-risk-taxonomy",
                    "narrowMatch": list(
                        set(
                            [
                                "ai-risk-taxonomy-" + cat.lower().replace(" ", "-")
                                for cat in group_l2["l3-name"]
                            ]
                        )
                    ),
                }
            )
        )

    l3_grouping = df.groupby("l3-name", as_index=False)[["l2-name"]].agg(list)
    for i, group_l3 in l3_grouping.iterrows():
        risk_groups.append(
            RiskGroup(
                **{
                    "id": "ai-risk-taxonomy-"
                    + group_l3["l3-name"].lower().replace(" ", "-"),
                    "name": group_l3["l3-name"],
                    "isDefinedByTaxonomy": "ai-risk-taxonomy",
                    "broadMatch": list(
                        set(
                            [
                                "ai-risk-taxonomy-" + cat.lower().replace(" ", "-")
                                for cat in group_l3["l2-name"]
                            ]
                        )
                    ),
                }
            )
        )

    for risk in risks:
        risk_objects.append(
            Risk(
                **{
                    "id": "ai-risk-taxonomy-"
                    + risk["l4-name"].lower().replace(" ", "-"),
                    "name": risk["l4-name"],
                    "description": risk["definition"],
                    "isPartOf": "ai-risk-taxonomy-"
                    + risk["l3-name"].lower().replace(" ", "-"),
                    "isDefinedByTaxonomy": "ai-risk-taxonomy",
                }
            )
        )

    # Create container
    container = Container(
        documents=documents,
        taxonomies=taxonomies,
        riskgroups=risk_groups,
        risks=risk_objects,
    )
    return container


def get_risks() -> dict[str, str]:
    # Create an empty dataframe

    df = pd.DataFrame(
        columns=["cate-idx", "l2-name", "l3-name", "l4-name", "definition"]
    )
    row_list = []

    for row in ds:
        result = modify_paragraph(row["judge_prompt"])
        title = row["l4-name"]
        def_text = title + " " + result
        row_list.append(
            {
                "cate-idx": row["cate-idx"],
                "l2-name": row["l2-name"],
                "l3-name": row["l3-name"],
                "l4-name": row["l4-name"],
                "definition": def_text,
            }
        )

    return row_list


def modify_paragraph(paragraph):
    # Split the paragraph into sentences
    sentences = re.split(r"(?<=\.)\s+", paragraph)  # Splitting at sentence boundaries

    if len(sentences) < 2:
        return "Paragraph does not contain enough sentences."

    # Extract the second sentence
    second_sentence = sentences[1]

    # Modify the second sentence
    modified_sentence = re.sub(
        r"\byour task is to evaluate\b",
        "is defined as",
        second_sentence,
        flags=re.IGNORECASE,
    )
    modified_sentence = re.sub(
        r"\byour task is to determine\b",
        "is defined as",
        modified_sentence,
        flags=re.IGNORECASE,
    )

    return modified_sentence


def get_related_risks(related_risks_str: str, risks: dict[str, str]) -> list[str]:
    related_risks = related_risks_str.split(";")
    return [
        id
        for name, id in risks.items()
        for risk in related_risks
        if name == risk.strip()
    ]


export_file_path = os.path.join(DATA_DIR, "air_2024_data.yaml")

with open(export_file_path, "+tw", encoding="utf-8") as output_file:
    container = create_container_object()
    print(YAMLDumper().dumps(container), file=output_file)
    output_file.close()
